from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals


class HashMismatch(ValueError):
    """
    Raised when the incoming hash of a file does not match the expected.
    """



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import re
import urlparse


class NormalizingDict(dict):

    def pop(self, key, default=None):
        value = super(NormalizingDict, self).pop(key, default)
        if not value or value in ["UNKNOWN", "None"]:
            value = default
        return value


_distutils2_version_capture = re.compile("^(.*?)(?:\(([^()]+)\))?$")


def split_meta(meta):
    meta_split = meta.split(";", 1)
    meta_name, meta_version = _distutils2_version_capture.search(meta_split[0].strip()).groups()
    meta_env = meta_split[1].strip() if len(meta_split) == 2 else ""

    return {
        "name": meta_name,
        "version": meta_version if meta_version is not None else "",
        "environment": meta_env,
    }


_url = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)


def clean_uri(url):
    parts = list(urlparse.urlsplit(url))

    if not parts[0]:
        # If no URL scheme given, assume http://
        parts[0] = "http"

    if not parts[1]:
        # Assume that if no domain is provided, that the path segment
        # contains the domain.
        parts[1] = parts[2]
        parts[2] = ""
        # Rebuild the url_fields list, since the domain segment may now
        # contain the path too.
        parts = list(urlparse.urlsplit(urlparse.urlunsplit(parts)))

    if not parts[2]:
        # the path portion may need to be added before query params
        parts[2] = "/"

    cleaned_url = urlparse.urlunsplit(parts)

    if not _url.search(cleaned_url):
        # Trivial Case Failed. Try for possible IDN domain
        if cleaned_url:
            scheme, netloc, path, query, fragment = urlparse.urlsplit(cleaned_url)

            try:
                netloc = netloc.encode("idna").decode("ascii")  # IDN -> ACE
            except UnicodeError:  # invalid domain part
                raise ValueError

            cleaned_url = urlparse.urlunsplit((scheme, netloc, path, query, fragment))

            if not _url.search(cleaned_url):
                raise ValueError
        else:
            raise ValueError

    return cleaned_url



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import collections
import datetime
import hashlib
import logging
import re
import time
import urlparse

from .pypi import Package


logger = logging.getLogger(__name__)


class Processor(object):

    def __init__(self, warehouse, pypi, store, *args, **kwargs):
        super(Processor, self).__init__(*args, **kwargs)

        self.warehouse = warehouse
        self.pypi = pypi
        self.store = store

    def get_and_update_or_create_version(self, release, project):
        version_data = release.serialize()
        version_data.update({"project": project})

        version, created = self.warehouse.versions.objects.get_or_create(project__name=project.name, version=release.version, show_yanked=True, defaults=version_data)

        if not created:
            version.classifiers = sorted(version.classifiers)

            changed = False

            for k, v in version_data.iteritems():
                if getattr(version, k, None) != v:
                    changed = True
                    setattr(version, k, v)

            if changed:
                version.save()

        return version

    def get_and_update_or_create_file(self, release, version, distribution):
        file_data = distribution.serialize()
        file_data.update({"version": version})

        vfile, created = self.warehouse.files.objects.get_or_create(filename=file_data["filename"], show_yanked=True, defaults=file_data)

        if not created:
            changed = False

            for k, v in file_data.iteritems():
                if getattr(vfile, k, None) != v:
                    changed = True
                    setattr(vfile, k, v)

            if changed:
                vfile.save()

        return vfile

    def update_files(self, release, version):
        # Determine if any files need to be deleted
        warehouse_files = set([f.filename for f in version.files])
        local_files = set([x.filename for x in release.files])
        deleted = warehouse_files - local_files

        # Delete any files that need to be deleted
        if deleted:
            for filename in deleted:
                logger.info("Deleting the file '%s' from '%s' version '%s'", filename, release.name, release.version)

            self.warehouse.files.objects.filter(filename__in=deleted).delete()

        return [self.get_and_update_or_create_file(release, version, distribution) for distribution in release.files]

    def update(self, name, version=None, timestamp=None, action=None, matches=None):
        package = Package(self.pypi, name, version)

        # Process the Name
        project, _ = self.warehouse.projects.objects.get_or_create(name=name)

        for release in package.releases():
            if "/" in release.version:
                # We cannot accept versions with a / in it.
                logger.error("Skipping '%s' version '%s' because it contains a '/'", release.name, release.version)
                continue

            if not release.changed(self.store.get("pypi:process:%s:%s" % (release.name, release.version))):
                logger.info("Skipping '%s' version '%s' because it has not changed", release.name, release.version)
                continue

            logger.info("Syncing '%s' version '%s'", release.name, release.version)

            version = self.get_and_update_or_create_version(release, project)
            self.update_files(release, version)

            self.store.set("pypi:process:%s:%s" % (release.name, release.version), release.hash())

    def delete(self, name, version, timestamp, action, matches):
        filename = None

        if action == "remove":
            if version is None:
                obj = self.warehouse.projects.objects.filter(name=name)
                logger.info("Deleting '%s'", name)
            else:
                obj = self.warehouse.versions.objects.filter(project__name=name, version=version)
                logger.info("Deleting '%s' version '%s'", name, version)
        elif action.startswith("remove file"):
            filename = matches.groups()[0]
            obj = self.warehouse.files.objects.filter(filename=filename)
            logger.info("Deleting '%s' version '%s' filename '%s'", name, version, filename)
        else:
            raise RuntimeError("Unknown Action passed to delete()")

        try:
            obj = obj.get()
        except obj.resource.DoesNotExist:
            return

        if version is None:
            key_pattern = "pypi:process:%s:*" % name
            keys = self.store.keys(key_pattern)
        else:
            keys = ["pypi:process:%s:%s" % (name, version)]

        for key in keys:
            self.store.delete(key)

        obj.delete()

    def process(self):
        logger.info("Starting changed projects synchronization")

        if not self.store.get("pypi:since"):
            # This is the first time we've ran so we need to do a bulk import
            raise RuntimeError(" Cannot process changes with no value for the last successful run.")

        current = datetime.datetime.utcnow()

        since = int(float(self.store.get("pypi:since"))) - 10

        dispatch = collections.OrderedDict([
            (re.compile("^create$"), self.update),
            (re.compile("^new release$"), self.update),
            (re.compile("^add [\w\d\.]+ file .+$"), self.update),
            (re.compile("^remove$"), self.delete),
            (re.compile("^remove file (.+)$"), self.delete),
            (re.compile("^update [\w]+(, [\w]+)*$"), self.update),
            #(re.compile("^docupdate$"), docupdate),  # @@@ Do Something
            #(re.compile("^add (Owner|Maintainer) .+$"), add_user_role),  # @@@ Do Something
            #(re.compile("^remove (Owner|Maintainer) .+$"), remove_user_role),  # @@@ Do Something
        ])

        changes = self.pypi.changelog(since)

        if changes:
            if isinstance(changes[0], basestring):
                changes = [changes]

        for name, version, timestamp, action in changes:
            action_hash = hashlib.sha512(":".join([str(x) for x in [name, version, timestamp, action]])).hexdigest()[:32]
            action_key = "pypi:changelog:%s" % action_hash

            logdata = {"action": action, "name": name, "version": version, "timestamp": timestamp}

            if not self.store.exists(action_key):
                logger.debug("Processing %(name)s %(version)s %(timestamp)s %(action)s" % logdata)

                # Dispatch Based on the action
                for pattern, func in dispatch.iteritems():
                    matches = pattern.search(action)
                    if matches is not None:
                        func(name, version, timestamp, action, matches)
                        break

                self.store.setex(action_key, 2592000, "1")
            else:
                logger.debug("Skipping %(name)s %(version)s %(timestamp)s %(action)s" % logdata)

        # Hijack the warehouse session and url
        last_modified_url = urlparse.urljoin(self.warehouse.url, "/last-modified")
        resp = self.warehouse.session.post(last_modified_url, {"date": current.isoformat()})
        resp.raise_for_status()

        self.store.set("pypi:since", time.mktime(current.timetuple()))

        logger.info("Finished changed projects synchronization")



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from .core import Carrier


def main():
    Carrier().run()


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import base64
import collections
import hashlib
import json

import requests

from .exceptions import HashMismatch
from .utils import NormalizingDict, clean_uri, split_meta


class File(object):

    def __init__(self, *args, **kwargs):
        kwargs = NormalizingDict(kwargs.items())

        # Useful data
        self.comment = kwargs.pop("comment_text")
        self.filename = kwargs.pop("filename")
        self.type = kwargs.pop("packagetype")
        self.python_version = kwargs.pop("python_version")
        self.created = kwargs.pop("upload_time")
        self.data = kwargs.pop("file")

        # PyPI internal data
        self._downloads = kwargs.pop("downloads")
        self._has_sig = kwargs.pop("has_sig")
        self._md5_digest = kwargs.pop("md5_digest")
        self._size = kwargs.pop("size")
        self._url = kwargs.pop("url")

        super(File, self).__init__(*args, **kwargs)

    def serialize(self):
        data = {
            "file": {
                "name": self.filename,
                "file": self.data,
            },
            "created": self.created,
            "type": self.type,
            "python_version": self.python_version,
            "comment": self.comment,
            "filename": self.filename,
            "filesize": self._size,
            "digests": {
                "md5": hashlib.md5(base64.b64decode(self.data)).hexdigest(),
                "sha256": hashlib.sha256(base64.b64decode(self.data)).hexdigest(),
            },
        }

        return data


class Release(object):

    def __init__(self, *args, **kwargs):
        kwargs = NormalizingDict(kwargs.items())

        # Useful data
        self.author = kwargs.pop("author", None)
        self.author_email = kwargs.pop("author_email", None)

        self.classifiers = kwargs.pop("classifiers", [])

        self.description = kwargs.pop("description", None)

        self.keywords = kwargs.pop("keywords", "")

        # Check for a comma
        if "," in self.keywords:
            self.keywords = [x.strip() for x in self.keywords.split(",")]
        else:
            self.keywords = [x.strip() for x in self.keywords.split()]

        self.license = kwargs.pop("license", None)

        self.maintainer = kwargs.pop("maintainer", None)
        self.maintainer_email = kwargs.pop("maintainer_email", None)

        self.name = kwargs.pop("name", None)

        self.platforms = kwargs.pop("platform", [])

        if isinstance(self.platforms, basestring):
            self.platforms = [self.platforms]

        self.supported_platforms = kwargs.pop("supported_platforms", [])

        if isinstance(self.supported_platforms, basestring):
            self.supported_platforms = [self.supported_platforms]

        self.requires_python = kwargs.pop("requires_python", None)

        self.summary = kwargs.pop("summary", None)

        self.version = kwargs.pop("version", None)

        self.uris = {}

        for key, label in {"bugtrack_url": "Bug tracker", "home_page": "Home page", "download_url": "Download", "docs_url": "Documentation"}.items():
            uri = kwargs.pop(key, None)

            if uri is not None:
                try:
                    self.uris[label] = clean_uri(uri)
                except ValueError:
                    pass

        for purl in kwargs.pop("project_url", []):
            label, uri = purl.split(",", 1)

            try:
                self.uris[label] = clean_uri(uri)
            except ValueError:
                pass

        self.requires = [split_meta(req) for req in kwargs.pop("requires_dist", [])]
        self.provides = [split_meta(req) for req in kwargs.pop("provides_dist", [])]
        self.obsoletes = [split_meta(req) for req in kwargs.pop("obsoletes_dist", [])]

        self.requires_external = kwargs.pop("requires_external", [])

        self._files = [File(**x) for x in kwargs.pop("files", [])]

        # Old and useless
        self._old_requires = kwargs.pop("requires", [])
        self._old_provides = kwargs.pop("provides", [])
        self._old_obsoletes = kwargs.pop("obsoletes", [])

        # PyPI internal data
        self._package_url = kwargs.pop("package_url", None)
        self._release_url = kwargs.pop("release_url", None)

        self._cheesecake_code_kwalitee_id = kwargs.pop("cheesecake_code_kwalitee_id", None)
        self._cheesecake_documentation_id = kwargs.pop("cheesecake_documentation_id", None)
        self._cheesecake_installability_id = kwargs.pop("cheesecake_installability_id", None)

        self._pypi_hidden = kwargs.pop("_pypi_hidden", None)
        self._pypi_ordering = kwargs.pop("_pypi_ordering", None)

        self._stable_version = kwargs.pop("stable_version", None)

        super(Release, self).__init__(*args, **kwargs)

    @property
    def files(self):
        return self._files

    def serialize(self):
        data = {}

        for key, value in self.__dict__.items():
            if key.startswith("_"):
                continue
            data[key] = value

        return data

    def hash(self):
        def _dict_constant_data_structure(dictionary):
            data = []

            for k, v in dictionary.items():
                if isinstance(v, dict):
                    v = _dict_constant_data_structure(v)
                elif isinstance(v, set):
                    v = sorted(v)
                data.append([k, v])

            return sorted(data, key=lambda x: x[0])

        data = self.serialize()
        data["files"] = [f.serialize() for f in self.files]
        data = json.dumps(_dict_constant_data_structure(data), default=lambda obj: obj.isoformat() if hasattr(obj, "isoformat") else obj)

        return hashlib.sha512(data).hexdigest()[:32]

    def changed(self, other):
        return not self.hash() == other


class Package(object):

    def __init__(self, client, package, version=None, *args, **kwargs):
        super(Package, self).__init__(*args, **kwargs)

        self.client = client
        self.package = package
        self.version = version

        self.session = requests.session()

    def versions(self):
        if self.version is None:
            versions = self.client.package_releases(self.package, True)

            if isinstance(versions, basestring):
                versions = [versions]
        else:
            versions = [self.version]

        return versions

    def releases(self):
        for version in self.versions():
            item = self.client.release_data(self.package, version)

            if not item:
                continue

            # fix classifiers
            item["classifiers"] = sorted(set(item.get("classifiers", [])))

            # Include the files
            item["files"] = self.files(version)

            yield Release(**item)

    def files(self, version):
        urls = self.client.release_urls(self.package, version)

        if isinstance(urls, collections.Mapping):
            urls = [urls]
        elif isinstance(urls, collections.Iterable):
            pass  # No action is required if it's already iterable
        else:
            raise ValueError("Do not understand the type returned by release_urls")

        files = []

        for url in urls:
            resp = self.session.get(url["url"], prefetch=True)
            resp.raise_for_status()

            file_hash = hashlib.md5(resp.content)

            if url["md5_digest"] != file_hash.hexdigest():
                raise HashMismatch("'MD5 hash {hash}' does not match the expected '{expected}' for {url}".format(hash=file_hash.hexdigest(), expected=url["md5_digest"], url=url["url"]))

            url["file"] = base64.b64encode(resp.content)

            files.append(url)

        return files



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import datetime
import logging
import time

from requests.exceptions import ConnectionError, HTTPError

from ..core import Carrier


logger = logging.getLogger(__name__)


# We ignore the last component as we cannot properly handle it
def get_jobs(last=0):
    current = time.mktime(datetime.datetime.utcnow().timetuple())

    print "Current time is", current

    app = Carrier()

    for package in set(app.processor.pypi.list_packages()):
        yield package


def handle_job(name):
    try:
        tried = 0
        delay = 1

        while True:
            try:
                tried += 1

                app = Carrier()
                app.processor.update(name)

                break
            except (ConnectionError, HTTPError):
                # Attempt to process again if we have a connection error
                if tried >= 10:  # Try a max of 10 times
                    raise
                else:
                    # Wait a moment
                    time.sleep(delay)
                    delay = delay * 2
    except Exception as e:
        logger.exception(str(e))
        raise



import os

SCHEDULE = {
    "packages": {"seconds": 30},
}

WAREHOUSE_URI = "https://api.crate.io/v1/"

PYPI_URI = "https://pypi.python.org/pypi"
PYPI_SSL_VERIFY = os.path.join(os.path.dirname(__file__), "pypi.crt")

REDIS = {}  # We leave this empty so client defaults occur

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,

    "formatters": {
        "default": {
            "format": "[%(asctime)s][%(levelname)s] %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },

    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "devnull": {
            "class": "logging.NullHandler",
        }
    },

    "root": {
        "handlers": ["console"],
        "level": "INFO",
    },

    "loggers": {
        "requests.packages.urllib3": {
            "handlers": ["devnull"],
            "propagate": False,
        }
    },
}



#!/usr/bin/env python
from setuptools import setup, find_packages

import carrier


install_requires = [
    "APScheduler",
    "forklift",
    "redis",
    "six",
    "xmlrpc2",
]

setup(
    name="carrier",
    version=carrier.__version__,

    description="Warehouse and PyPI Synchronization",
    long_description=open("README.rst").read(),
    url="https://github.com/crateio/carrier/",
    license=open("LICENSE").read(),

    author="Donald Stufft",
    author_email="donald.stufft@gmail.com",

    install_requires=install_requires,

    packages=find_packages(exclude=["tests"]),
    package_data={"": ["LICENSE"], "carrier": ["config/*.crt"]},
    zip_safe=False,

    entry_points={
        "console_scripts": [
            "carrier = carrier.__main__:main",
        ],
    },
)



from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import logging
import logging.config
import os
import time

import forklift
import redis
import requests
import xmlrpc2.client

from apscheduler.scheduler import Scheduler

from .config import Config, defaults
from .processor import Processor


logger = logging.getLogger(__name__)


class Carrier(object):

    def __init__(self, *args, **kwargs):
        super(Carrier, self).__init__(*args, **kwargs)

        # Get configuration
        self.config = Config(os.getcwd())
        self.config.from_object(defaults)

        if "CARRIER_CONF" in os.environ:
            self.config.from_envvar("CARRIER_CONF")

        # Initalize app
        logging.config.dictConfig(self.config["LOGGING"])

        store = redis.StrictRedis(**dict([(k.lower(), v) for k, v in self.config["REDIS"].items()]))

        wsession = requests.session(auth=(
                        self.config["WAREHOUSE_AUTH"]["USERNAME"],
                        self.config["WAREHOUSE_AUTH"]["PASSWORD"],
                    ))
        warehouse = forklift.Forklift(session=wsession)

        psession = requests.session(verify=self.config["PYPI_SSL_VERIFY"])
        ptransports = [xmlrpc2.client.HTTPTransport(session=psession), xmlrpc2.client.HTTPSTransport(session=psession)]
        pypi = xmlrpc2.client.Client(self.config["PYPI_URI"], transports=ptransports)

        self.processor = Processor(warehouse, pypi, store)

    def run(self):
        scheduler = Scheduler()

        if self.config["SCHEDULE"].get("packages") is not None:
            scheduler.add_interval_job(self.processor.process, **self.config["SCHEDULE"]["packages"])

        scheduler.start()

        try:
            while True:
                time.sleep(999)
        except KeyboardInterrupt:
            logger.info("Shutting down Carrier...")
            scheduler.shutdown(wait=False)
